#include "component_text.h"

void
component_text_glyph_size_get(ComponentText* self, Texture* texture, vec2 size)
{
	size[0] = texture->size[0] * self->scale[0];
	size[1] = texture->size[1] * self->scale[1];
}

void
component_text_glyph_draw(ComponentText* self, ECS* ecs, Texture* texture, vec3 position)
{
	mat4 model;
	vec2 size;

	component_text_glyph_size_get(self, texture, size);

	glm_mat4_identity(model);
	glm_translate(model, position);

	texture_quad_draw
	(
	 	texture,
	 	ecs->renderer,
		self->shader, 
		(f32*)TEXTURE_QUAD_UV_MIN_DEFAULT,
		(f32*)TEXTURE_QUAD_UV_MAX_DEFAULT,
		model, 
		size, 
		self->color,
		FLIP_NONE
	);
}

void
component_text_rectangle_get(ComponentText* self, Rectangle* rectangle)
{
	char glyph;
	char nextGlyph;
	f32 lineW;
	f32 offsetX;
	f32 prevY;
	GlyphMetrics glyphMetrics;
	vec2 origin;
	vec2 size;
	f32 lineSkip;
	f32 kerning;
	f32 advance;

	lineSkip = font_line_skip_get(self->font) * self->scale[1];

	font_glyph_metrics_init(self->font, &glyphMetrics, self->string[0]);

	lineW = 0.0f;

	size[0] = 0.0f;
	size[1] = (f32)glyphMetrics.maxY - glyphMetrics.minY;

	for (s32 i = 0; i < (s32)self->length; i++)
	{
		glyph = self->string[i];

		if 
		(
			glyph == '\n' ||
			(glyph == ' ' && (lineW > self->wrap) && self->wrap != -1)
		)
		{
			size[1] += lineSkip;
			lineW = 0.0f;

			if (glyph == '\n')
				continue;
		}

		if (i < (s32)self->length)
		{
			nextGlyph = self->string[i + 1];
			kerning = (f32)font_glyph_kerning_get(self->font, glyph, nextGlyph);
		}
		else
			kerning = 0.0;

		font_glyph_metrics_init(self->font, &glyphMetrics, glyph);

		advance = glyphMetrics.advance;

		offsetX = advance + kerning;
		offsetX *= self->scale[0];
	
		lineW += offsetX;

		if (lineW > size[0])
			size[0] = lineW;
	}

	rectangle->w = size[0];
	rectangle->h = size[1];
	rectangle->x = self->position[0] + self->offset[0];
	rectangle->y = self->position[1] + self->offset[1];
}

void
component_text_string_set(ComponentText* self, const char* string, u32 length)
{
	if (length > COMPONENT_TEXT_STRING_MAX)
		return;

	self->length = length;

	memset(self->string, '\0', sizeof(char) * COMPONENT_TEXT_STRING_MAX);
	memcpy(self->string, string, sizeof(char) * self->length);
}

void
component_text_add(ComponentText* self, ECS* ecs)
{
	glm_vec4_copy(COMPONENT_TEXT_COLOR_DEFAULT, self->color);
	glm_vec2_copy((f32*)COMPONENT_TEXT_SCALE_DEFAULT, self->scale);
}

void
component_text_draw(ComponentText* self, ECS* ecs)
{
	char glyph;
	char nextGlyph;
	f32 advance;
	f32 lineW;
	f32 offsetX;
	f32 kerning;
	GlyphMetrics glyphMetrics;
	Rectangle rectangle;
	vec2 origin;
	vec3 position;
	f32 lineSkip;
	f32 lineBegin;

	component_text_rectangle_get(self, &rectangle);

	lineSkip = font_line_skip_get(self->font) * self->scale[1];

	glm_vec3_zero(position);

	position[0] = rectangle.x;
	position[1] = rectangle.y;

	position[0] += self->offset[0];
	position[1] += self->offset[1];
	position[2] = self->position[2] + self->offset[2];

	offsetX = 0.0f;
	lineW = 0.0f;

	lineBegin = position[0];

	for (s32 i = 0; i < (s32)self->length; i++)
	{
		glyph = self->string[i];
	
		if 
		(
			glyph == '\n' ||
			(glyph == ' ' && (lineW > self->wrap) && self->wrap != -1)
		)
		{
			position[0] = lineBegin;
			position[1] += lineSkip;

			lineW = 0.0f;

			if (glyph == '\n')
				continue;
		}

		if (i < (s32)self->length)
		{
			nextGlyph = self->string[i + 1];
			kerning = (f32)font_glyph_kerning_get(self->font, glyph, nextGlyph);
		}
		else
			kerning = 0.0;

		font_glyph_metrics_init(self->font, &glyphMetrics, glyph);

		component_text_glyph_draw(self, ecs, &self->font->glyphTextures[(s32)glyph], position);

		advance = glyphMetrics.advance;

		offsetX = advance + kerning;
		offsetX *= self->scale[0];

		position[0] += offsetX;
		lineW += offsetX;
	}
}

void
component_text_init
(
	ComponentText* self, 
	ECS* ecs,
	Font* font,
	Shader* shader,
	RendererBuffer buffer,
	const char* string,
	const vec2 scale,
	const vec3 position,
	const vec3 offset,
	const vec4 color,
	u32 length,
	s32 wrap
)
{
	self->font = font;
	self->wrap = wrap;
	self->shader = shader;
	self->buffer = buffer;

	glm_vec2_copy((f32*)scale, self->scale);
	glm_vec3_copy((f32*)position, self->position);
	glm_vec3_copy((f32*)offset, self->offset);
	glm_vec4_copy((f32*)color, self->color);

	component_text_string_set(self, string, length);
}
